;;; "ff.scm" Polynomial factorization.		-*-scheme-*-
;;; Copyright 1994, 1995 Mike Thomas
;;; See the file "COPYING" for terms applying to this program.

;;; AUTHOR
;;;	  Mike Thomas
;;;	  46 Harold Street
;;;	  STAFFORD  QLD	 4053
;;;	  AUSTRALIA

;;;	  Email: mjt@octavia.anu.edu.au

;;;	  Phone: Intl + 61 7 356 8494

;;; SOURCES
;;; These functions are drawn from Geddes et al, 'Algorithms for
;;; Computer Algebra', Kluwer, 1992 (GCL), Davenport et al, and Knuth
;;; 'Seminumerical Algorithms'.

;;; DEVELOPMENT LANGUAGE
;;; SCM 4e1, compiled with DICE v 2.07.54

;;; DEVELOPMENT SYSTEM
;;; Amiga 1200, 2Mb Chip RAM, 4Mb Fast RAM, 50MHz 68030 CPU No FPU,
;;; Workbench 3.0

;;; DEPENDENCIES
;;; You need R4RS Scheme, SLIB and JACAL (latest versions).
;;; ff:p+p and ff:p-p depend on the behaviour of map when the argument
;;;    lists are of different lengths.
;;; Multiargument *, +
;;; You need array-copy! (not in slib yet)

;;; REPRESENTATIONS
;;; The modular homomorphism sym:sym maps the integers to a symmetric
;;; representation, following the numerous examples in GCL. This
;;; mapping does not seem to correspond with GCL's own definition of Z
;;; mod n using a simple remainder algorithm.

;;; The univariate polynomials expected and returned are JACAL's list
;;; representation, and are never numbers by themselves.
;;; There is no attempt to cover sparse polynomials yet.

;;; An ideal is a list of lists specifying a variable and then it's
;;; value eg '((y 1) (z 2) (p -1))

;;; NAMING
;;;	     u: univariate polynomials, should be JACAL's univ:
;;;	    ff: polynomials over finite integer fields using the
;;;		symmetic representation
;;;	   sym: symmetric representation integers
;;;	  poly: Jacal poly types
;;;	   hen: Hensel lifting procedures

;;; ARGUMENT CODING CONVENTIONS
;;;	    ff: the modulus argument is the last in the argument list

;;; KNOWN BUGS

;;; -- Output format gets misaligned, possibly JACAL's problem. The
;;;    factorisation stuff outputs a list of n occurrences of (factor
;;;    exponent)

;;; -- No attempt is made to use primes larger than those on the list
;;;    *primes*.  I intend to change this eventually, although for
;;;    small computers, at present, such large primes will cost a lot
;;;    of run time.

;;; -- The implementation of Hensel lifting is naive and therefore
;;;    slow.

;;; -- Multivariate Hensel lifting and factorisation is unfinished and
;;;    only just into the debugging stage.

;;; HISTORY moved to ChangeLog

(require 'array)
(require 'common-list-functions)
(require 'modular)
(require 'hash-table)
(require 'sort)

;;; (jaffer) printer for diagnostic information
(define (ff:print . args)
  (define result #f)
  (for-each (lambda (x) (set! result x) (display-diag x) (display #\ )) args)
  (newline-diag)
  result)

;;; Memoize an n argument procedure fn from Norvig 'artificial
;;; Intelligence Programming'
;;; Examples
;;; (define fib
;;;   (memon (lambda (n)
;;;	 (if (<= n 1) 1
;;;	    (+ (fib (- n 1)) (fib (- n 2)))))))

;;; (define sum
;;;   (memon (lambda (x y)
;;;	 (+ x y))))

(define (memon fn)
  (let ((table (make-hash-table 100))
	(gethash (hash-inquirer equal?))
	(puthash (hash-associator equal?)))
    (lambda (. x)
      (let ((val (gethash table x)))
	(if val
	    val
	    (let ((fx (apply fn x)))
	      (puthash table x fx)
	      fx))))))

;;; Define the symmetric modular homomorphism from a in Z to Z mod o,
;;; o > 0 is odd. This needs to be written in C, or a better algorithm
;;; substituted. See the benchmark data at the end for the effect of
;;; converting to this representation - mainly due to sym:sym I
;;; suspect. It is used in all the low level procedures.  I tried
;;; memoising, but it is slower.

;;;(define sym:symm
;;; (memon (lambda (a o)
;;;    (let ((o2 (/ (- o 1) 2))
;;;	     (a1 (remainder a o)))
;;;	  (if (> (abs a1) o2)
;;;	     (if (negative? a1)
;;;		(+ a1 o)
;;;		(- a1 o))
;;;	     a1)))))

(define (sym:sym a o)
  (let ((a1 (remainder a o)))
    (if (> (abs a1) (/ (- o 1) 2))
	(if (negative? a1)
	    (+ a1 o)
	    (- a1 o))
	a1)))

;;; find the bound of Z mod o, where o must be odd
(define (sym:bound o)
  (/ (- o 1) 2))

(define (ff:degree p n)
  (- (length (ff:unorm p n)) 2))

(define (u:degree p)
  (- (length p) 2))

(define (ff:leading p n)
  (sym:sym (car (last-pair p)) n))

(define (u:leading p)
  (car (last-pair p)))

(define u:var car)

(define (u:zero? p)
  (and (= (length p) 2) (zero? (cadr p))))

(define (u:one? p)
  (and (= (length p) 2) (= (cadr p) 1)))

(define (u:number? p)
  (and (= (length p) 2) (number? (cadr p))))

(define (poly:poly? p)
  (cond
    ((not (list?  p)) #f)
    ((not (poly:var? (car p))) #f)
    (else #t)))

(define (u:univariate? p)
  (cond
    ((not (poly:poly?  p)) #f)
    ((not (= (length (poly:vars e)) 1)) #f)
    (else #t)))

;;; remove duplicates from a list
(define (remove-duplicates l l1)
  (cond ((null? l) l1)
	((member (car l) l1) (remove-duplicates (cdr l) l1))
	(else (remove-duplicates (cdr l) (cons (car l) l1)))))

;;; convert a zero degree univariate poly to a number
(define (u:convert p)
  (if (u:number? p)
      (cadr p)
      p))

;;; Return a polynomial in variable v if given a number, otherwise the
;;; polynomial.  Assumes "otherwise" is a polynomial
(define (number->poly a v)
  (if (number? a)
      (list v a)
      a))

(define *primes*
  '(3 5 7 11 13 17 19 23 29 31 37 41 43 47 53 59 61
      67 71 73 79 83 89 97 101 103 107 131 137 149))

(define (sign a)
  (if (negative? a) -1 1))

;;; Extended Euclidean algorithm returns the list (g s t),
;;; where g is the gcd of a and b, and g = s*a + t*b.
;;; GCL algorithm 2.2

(define (sym:eea-gcd a b n)
  (letrec ((feg (lambda (c c1 c2 d d1 d2)
		  (if (not (zero? d))
		      (let ((q (sym:sym (quotient c d) n)))
			(feg d
			     d1
			     d2
			     (sym:sym (- c (* q d)) n)
			     (sym:sym (- c1 (* q d1)) n)
			     (sym:sym (- c2 (* q d2)) n)))
		      (list c c1 c2)))))
    (let* ((result (feg (abs a) 1 0 (abs b) 0 1))
	   (c (car result)))
      (list (sym:sym (abs c) n)
	    (sym:sym (* (cadr result) (sign a) (sign c)) n)
	    (sym:sym (* (caddr result) (sign b) (sign c)) n)))))

;;; and also for poly's over Z mod n
(define (ff:eea-gcd a b n)
  (letrec ((feg (lambda (c c1 c2 d d1 d2)
		  (if (not (u:zero? d))
		      (let* ((qr (ff:p/p->qr c d n))
			     (q	 (car qr))
			     (r1 (ff:p-p c1 (ff:p*p q d1 n) n))
			     (r2 (ff:p-p c2 (ff:p*p q d2 n) n)))
			(feg d d1 d2 (cadr qr) r1 r2))
		      (list c c1 c2)))))
    (let* ((result (feg (u:pp a) (list (u:var a) 1) (list (u:var a) 0)
			(u:pp b) (list (u:var a) 0) (list (u:var a) 1)))
	   (c (car result)))
      (list (ff:make-monic c n)
	    (ff:p/c (cadr result) (* (ff:leading a n) (ff:leading c n)) n)
	    (ff:p/c (caddr result)(* (ff:leading b n) (ff:leading c n)) n)))))

(define (poly:eea-gcd a b)
   (letrec ((feg (lambda (c c1 c2 d d1 d2)
		 (if (not (poly:0? d))
		    (let* ((qr (poly:pdiv c d))
			   (q  (car qr))
			   (r1 (poly:- c1 (poly:* q d1)))
			   (r2 (poly:- c2 (poly:* q d2))))
		       (feg d d1 d2 (cadr qr) r1 r2))
		       (list c c1 c2)))))
      (let* ((result (feg (u:pp a) (list (u:var a) 1) (list (u:var a) 0)
			  (u:pp b) (list (u:var a) 0) (list (u:var a) 1))))
	result)))

;
;(c (car result)))
;	  (list c
;		(ff:p/c (cadr result) (* (ff:leading a n) (ff:leading c n)) n)
;		(ff:p/c (caddr result)(* (ff:leading b n) (ff:leading c n)) n)))))


;;; Standard Euclidean algorithm for gcd
(define (ff:euclid-gcd a b n)
  (let ((r 0))
    (let loop ((c (u:pp (ff:unorm a n)))
	       (d (u:pp (ff:unorm b n))))
      (if (not (u:zero? d))
	  (begin
	    (set! r (cadr (ff:p/p->qr c d n)))
	    (loop d (u:pp r)))
	  (ff:make-monic (ff:p*c c (gcd (u:cont a) (u:cont b)) n) n)))))

;;; based on modular:invert, but arguments swapped to conform with my
;;; conventions.

;;;(define sym:invertm
;;;(memon (lambda (a m)
;;;  (let* ((d1 (abs a))
;;;	    (d (modular:extended-euclid d1 m)))
;;;    (if (= 1 (car d))
;;;	   (* (sign a) (sym:sym (cadr d) m))
;;;	   (slib:error "sym:invert can't invert" a m))))))

(define (sym:invert a m)
  (let ((d (modular:extended-euclid (abs a) m)))
    (if (= 1 (car d))
	(* (sign a) (sym:sym (cadr d) m))
	(slib:error "sym:invert can't invert" a m))))

;;;; Find a number x such that (x * s) mod p = r
;;;(define sym:multm (memon (lambda (r s p)
;;; (if (zero? (sym:sym r p))
;;;    0
;;;    (if (memv p *primes*)
;;;	  (let ((m (sym:sym r p))
;;;		(n (sym:sym s p))
;;;		(last (sym:bound p))
;;;		(mlast (* (sym:bound p) -1)))
;;;	       (do ((i mlast (+ i 1)))
;;;		   ((or (>= i last)
;;;			(= (sym:sym (* i n) p) m))
;;;		    i)))
;;;	  (error
;;;	     "In sym:mult ; Third argument is not on my list of primes" p))))))

(define (sym:mult r s p)
  (if (zero? (sym:sym r p))
      0
      (if (memv p *primes*)
	  (let ((m (sym:sym r p))
		(n (sym:sym s p))
		(last (sym:bound p))
		(mlast (* (sym:bound p) -1)))
	    (do ((i mlast (+ i 1)))
		((or (>= i last) (= (sym:sym (* i n) p) m)) i)))
	  (error
	   "In sym:mult ; Third argument is not on my list of primes" p))))

(define (ff:monic? p n)
  (if (not (u:number? p))
      (= (ff:leading p n) 1)
      #f))

(define (ff:make-monic p n)
  (if (not (ff:monic? p n))
      (ff:p*c p (sym:invert (ff:leading p n) n) n)
      p))

;;; Numeric content and primitive part of a univariate polynomial p
(define (u:cont p)
  (apply gcd (cdr p)))

(define (u:pp p)
  (if (u:number? p)
      p
      (let ((c (u:cont p)))
	(cons (u:var p)
	      (map (lambda (x) (/ x c)) (cdr p))))))

;;; Polynomial content and primitive part of a multivariate polynomial p
;;; with respect to var.
(define (poly:cont p var)
  (let loop ((deg (- (poly:degree p var) 1))
	     (cgcd (leading-coeff p var)))
;;;    (ff:print "deg = " deg " cgcd = " cgcd)
    (if (>= deg 0)
      (loop (- deg 1) (let ((coeft (poly:coeff p var deg)))
			(if (number? coeft)
			  (if (zero? coeft)
			    cgcd
			    (poly:gcd cgcd coeft))
			  (poly:gcd cgcd coeft))))
      cgcd)))

(define (poly:pp p var)
  (if (number? p)
      p
      (poly:/ p (poly:cont p var))))

;;; remove leading zeros of a poly with already modularised coefficients
(define (u:lose-lzeros p)
  (let* ((l  (length p))
	 (nl (do ((i  (- l 1) (- i 1)))
		 ((or (not (zero? (list-ref p i))) (< i 2)) (+ i 1)))))
    (if (not (= nl l))
	(reverse (list-tail (reverse p) (- l nl)))
	p)))

;;; reduce univariate p mod n and remove any consequential leading
;;; zero coefficients
(define (ff:unorm p n)
  (u:lose-lzeros (cons (u:var p) (map (lambda (x) (sym:sym x n)) (cdr p)))))

;;; Same but multivariate, copied from Jaffer's poly-mod-number in poly.scm
;;; and using sym:sym instead of modulo.
(define (ff:mnorm poly modulus)
  (if (number? poly)
      (sym:sym poly modulus)
    (cons (car poly)
	  (map-no-end-0s (lambda (x) (ff:mnorm x modulus))
			 (cdr poly)))))

;;; Add two numerical coefficient univariate polynomials mod n.
;;; Returns normalised sum.
(define (ff:p+p a b n)
  (let ((la (length a))
	(lb (length b)))
    (if (equal? (u:var a) (u:var b))
	(u:lose-lzeros
	 (cons (u:var a)
	       (append
		(map (lambda (x y) (sym:sym (+ x y) n))
		     (cdr a) (cdr b))
		(if (< la lb)
		    (map (lambda (x) (sym:sym x n)) (list-tail b la))
		    (map (lambda (x) (sym:sym x n)) (list-tail a lb))))))
	(error "In ff:p+p ; The arguments have different variables."
	       (u:var a)
	       (u:var b)))))

(define (ff:p-p a b n)
  (let ((la (length a))
	(lb (length b)))
    (if (equal? (u:var a) (u:var b))
	(u:lose-lzeros
	 (cons (u:var a)
	       (append
		(map (lambda (x y) (sym:sym (- x y) n))
		     (cdr a) (cdr b))
		(if (< la lb)
		    (map (lambda (x) (sym:sym (- 0 x) n)) (list-tail b la))
		    (map (lambda (x) (sym:sym x n)) (list-tail a lb))))))
	(error "In ff:p-p ; The arguments have different variables."
	       (u:var a)
	       (u:var b)))))

(define (ff:p*c a c n)
  (if (number? c)
      (u:lose-lzeros
       (cons (u:var a)
	     (map (lambda (x) (sym:sym (* x c) n))
		  (cdr a))))
      (error "In ff:p*c ; The constant should be a number." c)))

(define (ff:p*p a b n)
  (ff:unorm (number->poly (poly:* a b) (u:var a)) n))

(define (u:p*c a c)
  (let* ((p (list->vector a))
	 (lp (vector-length p)))
    (if (number? c)
	(do ((i 1 (+ i 1)))
	    ((>= i lp) (vector->list p))
	  (vector-set! p i (* (vector-ref p i) c)))
	(error "In ff:p*c ; The constant should be a number." c))))

(define (ff:p/c a c n)
  (if (number? c)
      (let ((ci (sym:invert c n)))
	(u:lose-lzeros
	 (cons (u:var a)
	       (map (lambda (x) (sym:sym (* x ci) n))
		    (cdr a)))))
      (error "In ff:p/c ; The constant should be a number." c)))

;;; Uses quotient not /
(define (u:p/c a c)
  (if (number? c)
      (cons (u:var a) (map (lambda (x) (quotient x c)) (cdr a)))
      (error "In u:p/c ; The constant should be a number." c)))

;;; promote the degree of a polynomial a to n
;;; assumes a is normalised
(define (u:higher-deg a n)
  (let ((dp (- (length a) 2)))
    (cond
     ((> n dp)
      (cons
       (u:var a)
       (append (vector->list (make-vector (- n dp) 0)) (cdr a))))
     ((= n dp) a)
     (else
      (error "In u:higher-deg ; The polynomial has higher degree than " n)))))

(define (ff:prem p q n)
  (let* ((dp (ff:degree p n))
	 (dq (ff:degree q n))
	 (lp (ff:leading p n))
	 (lq (ff:leading q n))
	 (g (gcd lp lq)))
    (cond ((u:zero? p) (list (u:var p) 0))
	  ((u:number? p) p)
	  ((u:number? q) (list (u:var p) 0))
	  ((< dp dq)
	   (ff:unorm p n))
	  (else (let* ((xq1 (u:higher-deg (ff:unorm q n) dp))
		       (xq2 (ff:p*c xq1
				    (sym:mult lp lq n)
				    n))
		       (r (ff:p-p p xq2 n)))
;;;		  (ff:print xq1 xq2 r)
		  (if (< (ff:degree r n) dq)
		      (ff:unorm r n)
		      (ff:prem r q n)))))))

;;; Returns the quotient and remainder of a and b over Z mod p, Leaves
;;; all error checks and argument normalisation to calling procedure
(define (ff:p/p->qr u v p)
  (let ((m (ff:degree u p))
	(n (ff:degree v p))
	(lu (ff:leading u p))
	(lv (ff:leading v p)))
    (cond
     ((u:zero? u) (list (list (u:var v) 0) (list (u:var v) 0)))
     ((equal? u v) (list (list (u:var v) 1) (list (u:var v) 0)))
     ((and (u:number? v) (u:number? u))
      (list (list (u:var v) (* (sym:invert (u:convert v) p) (u:convert u)))
	    (list (u:var v) 0)))
     ((u:number? v) (list (ff:p/c u (u:convert v) p) (list (u:var v) 0)))
     ((< m n)
      (list (list (u:var u) 0) u))
     (else
      (let ((multiplier (sym:invert lv p))
	    (uv (list->vector (cdr u)))
	    (vv (list->vector (cdr v)))
	    (q	(make-vector (+ (- m n) 1) 0))
	    (var (u:var u)))
	(let loop ((k (- m n)))
	  (if (negative? k)
	      (list (ff:unorm (cons var (vector->list q)) p)
		    (ff:unorm (cons var
				   (reverse (list-tail
					     (reverse (vector->list uv))
					     (+ 1 (- m n))))) p))
	      (let ((qk (* (vector-ref uv (+ n k)) multiplier)))
		(vector-set! q k qk)
		(let loop1 ((j (- (+ n k) 1)))
		  (if (>= j k)
		      (begin
			(vector-set! uv j (- (vector-ref uv j)
					     (* qk
						(vector-ref vv (- j k)))))
			(loop1 (- j 1)))))
		(loop (- k 1))))))))))

(define (ff:p/p a b p)
  (let ((u (ff:unorm a p))
	(v (ff:unorm b p)))
    (let ((m (ff:degree u p))
	  (n (ff:degree v p)))
      (cond
       ((and (u:zero? u) (u:zero? v))
	(error "In ff:p/p ; 0/0 is undefined."))
       ((u:zero? v)
	(error "In ff:p/p ; Division by 0 is undefined."))
       (else
	(car (ff:p/p->qr u v p)))))))

;;; Given a primitive polynomial a(v) which is an element of a
;;; Unique Factorisation Domain in v, calculate the square free
;;; factorisation of a(v). A primitive polynomial has the content
;;; removed from its coefficients.  The content is the gcd of the
;;; coefficients.
;;; from page 340 of Geddes et al, Algorithms for Computer Algebra
;;;    - mjt
(define (poly:sff a)
  (let* ((v (u:var a))
	 (b (poly:diff a v))
	 (c (signcan (poly:gcd a b)))
	 (w (poly:/   a c))
	 (y 0)
	 (z 0)
	 (output '()))
    (poly:sort-factors
      (let loop ((i 1))
	(if (not (equal? c 1))
	    (begin
		   (set! y (signcan (poly:gcd w c)))
		   (set! z (poly:/ w y))
;		    (ff:print "y = " y " w = " w " c = " c " z = " z)
		   (if (not (number? z))
		       (set! output (if (null? output)
					(list (list z i))
					(append output (list (list z i))))))
		   (set! w y)
		   (set! c (poly:/ c y))
;		    (ff:print "c = " c	)
		   (loop (+ i 1)))
	    (append output (list (list w i))))))))

(define (ff:gfroot p var n)
  (let ((d (ff:degree p n)))
    (if (zero? (sym:sym d n))
	(let* ((u (/ d n))
	       (v (make-vector (+ u 1) 0)))
	  (vector-set! v u 1)
	  (vector-set! v 0 1)
	  (cons var (vector->list v)))
	(error "In ff:gfroot ; The polynomial has no root. " p))))

(define (ff:diff p v n)
  (if (equal? (u:var p) v)
      (do ((i (- (length p) 1) (- i 1))
	   (r '() (cons (sym:sym (* (- i 1) (list-ref p i)) n) r)))
	  ((< i 2) (cons v r)))
      (list v 0)))

(define (ff:sff a p)
  (let* ((v (u:var a))
	 (b (ff:diff a v p))
	 (output '()))
;(ff:print "a = " a "  b = " b)
    (if (not (equal? b (list v 0)))
	(let* ((c (ff:euclid-gcd a b p))
	       (w (ff:p/p a c p))
	       (y 0)
	       (z 0))
	  (let loop ((i 1))
;(ff:print "i = " i "  c = " c)
	    (if (not (equal? w (list v 1)))
		(begin (set! y (ff:euclid-gcd w c p))
		       (set! z (ff:p/p w y p))
		       (if (not (u:one? z))
			   (set! output (if (null? output)
					    (list (list z i))
					    (append output
						    (list (list z i))))))
;(ff:print "y = " y "z = " z "output = " output)
		       (set! w y)
		       (set! c (ff:p/p c y p))
		       (loop (+ i 1)))))
;(ff:print "c = " c)
	  (if (not (equal? c (list v 1)))
	      (begin (set! c (ff:gfroot c v p))
		     (set! output (append output
					  (list (list (ff:sff c v p) p)))))))
	(begin (set! a (ff:gfroot a v p))
	       (set! output (append output (list (list (ff:sff a v p) p))))))
;(ff:print "output = " output)
    output))

(define (ff:q-matrix a p)
  (let* ((n (ff:degree a p))
	 (n-1 (- n 1))
	 (q (make-array 0 n n))
	 (r (make-vector n 0))
	 (r1 (make-vector n 0))
	 (u (* n-1 p))
	 (b (list->vector a)))
    (array-set! q 1 0 0)
    (vector-set! r 0 1)
    (vector-set! r1 0 1)
    (let loop ((i 1))
      (if (> i u)
	  q
	  (begin
	    (vector-set! r 0 (* (vector-ref b 1) (- 0 (vector-ref r1 n-1))))
	    (let loop1 ((j 1))
	      (if (<= j n-1)
		  (begin
		    (vector-set!
		     r j (sym:sym (- (vector-ref r1 (- j 1))
				     (sym:sym (* (vector-ref b (+ j 1))
						 (vector-ref r1 n-1)) p)) p))
		    (loop1 (+ j 1)))))
	    (array-map! r1 (lambda (x) x) r)
	    (if (zero? (modulo i p))
		(array-map! (make-shared-array
			     q (lambda (j) (list (/ i p) j)) n)
			    (lambda (x) x) r))
;;;		(ff:print i "  r = " r "  q = " q)
	    (loop (+ i 1)))))))

;;; Knuth's null-space-basis, slow but works
(define (ff:null-space-basis mo var p)
  (let* ((n (car (array-dimensions mo)))
	 (m (make-array #f n n))
	 (c (make-vector n -1))
	 (ivec (index-vector n))
	 (result ()))
;;;    (ff:print " mo = " mo " c = ")
    (array-copy! mo m)
    (do ((k 0 (+ k 1)))
	((>= k n))
      (let ((j (do ((b 0 (+ b 1)))
		   ((or (>= b n)
			(and (negative? (vector-ref c b))
			     (not (zero? (array-ref m k b))))) b))))
;;;	   (ff:print "j = " j " n = " n)
	(if (< j n)
	    (let ((muinv  (* -1 (sym:invert (array-ref m k j) p)))
		  (mcolj (make-shared-array m (lambda (a) (list a j)) n)))
;;;		 (ff:print "u = " u " muinv = " muinv)
	      (array-map! mcolj (lambda (x) (sym:sym (* x muinv) p)) mcolj)
	      (vector-set! c j k)
	      (do ((i 0 (+ i 1)))
		  ((>= i n))
		(if (not (= i j))
		    (let ((mcoli (make-shared-array
				  m (lambda (a) (list a i)) n))
			  (mki (array-ref m k i)))
		      (array-map! mcoli
				  (lambda (x y z)
				    (if (>= z k)
				      (sym:sym (+ x (* mki y)) p)
				      x))
				  mcoli mcolj ivec)))))
;;;		   (ff:print " m after col mults = " m)
;;;		   (ff:print " c after col mults = " c)
	    (let ((vr (make-vector n 0))
		  (cl (vector->list c)))
	      (array-map! vr
			  (lambda (i)
			    (let ((cl1 (memv i cl)))
			      (cond
				((= i k) 1)
				(cl1 (array-ref m k (- n (length cl1))))
				(else 0))))
			  ivec)
	      (set! result (append result
				   (list (u:lose-lzeros
					   (cons var (vector->list vr))))))
;;;		 (ff:print "c = " c " vr = " vr " result = " result)
	      ))))
    result))


;;; -- > (u:factorz p7)
;     ERROR: vector-ref: Argument out of range 2
;     ;Evaluation took 1934 mSec (601 in gc) 19181 cells work, 6439 bytes other
;     > p7
;     ;Evaluation took 1 mSec (0 in gc) 2 cells work, 31 bytes other
;     (x 1 -3 -1 -3 1 -3 1)
;     > (math)
;     type qed; to return to scheme, type help; for help.
;     e5 : factoruz(1-3*x-x^2-3*x^3+x^4-3*x^5 +x^6);
;     poly = (#(x x #f () #f ()) 1 -3 -1 -3 1 -3 1)
;     e1 = (#(x x #f () #f ()) 1 -3 -1 -3 1 -3 1)
;     Arithmetic Error; Last expression lost.

;;; This problem is caused by berlekamp expecting to find factors when
;;; there are none (ie the polynomial is irreducible.  This problem
;;; comes from the ff:null-space-basis-gcl-bug procedure below.  This
;;; is the only occurrence of this problem I am aware of.
(define (ff:null-space-basis-gcl-bug mo var p)
  (let* ((n (car (array-dimensions mo)))
	 (m (make-array #f n n)))
;;;    (ff:print " mo = " mo)
    (array-copy! mo m)
    (do ((k 0 (+ k 1)))
	((>= k n))
      (let ((i (do ((b k (+ b 1)))
		   ((or (>= b n) (not (zero? (array-ref m k b)))) b))))
	(if (< i n)
	    (let* ((u	  (array-ref m k i))
		   (uinv  (sym:invert u p))
		   (mcoli (make-shared-array m (lambda (a) (list a i)) n))
		   (mcolk (make-shared-array m (lambda (a) (list a k)) n))
		   (temp  (make-shared-array (make-array #f n n)
					     (lambda (a) (list a 0)) n)))
	      (array-map! mcoli (lambda (x) (sym:sym (* x uinv) p)) mcoli)
	      (array-map! temp	(lambda (x) x) mcoli)
	      (array-map! mcoli (lambda (x) x) mcolk)
	      (array-map! mcolk (lambda (x) x) temp)
	      (do ((j 0 (+ j 1)))
		  ((>= j n))
		(if (not (= j k))
		    (let ((mcolj (make-shared-array
				  m (lambda (a) (list a j)) n))
			  (mkj (array-ref m k j)))
		      (array-map! mcolj
				  (lambda (x y)
				    (sym:sym (- x (sym:sym (* mkj y) p)) p))
				  mcolj mcolk))))
;;;(ff:print " m = " m)
	      ))))
    (let ((mdiag (make-shared-array m (lambda (a) (list a a)) n)))
      (array-map! mdiag (lambda (x) (sym:sym (- x 1) p)) mdiag))
;;;    (ff:print "  After subtraction of 1 along the diagonal,	m = " m))
;;;	 (array-map! m (lambda (x) (sym:sym (* x -1) p)) m)
;;;    (ff:print "  After multiplication of -1 along the diagonal,  m = " m)
    (let ( ;(i 0)
	  (ret '()))
      (let loop1 ((j 0))
	(if (< j n)
	    (let ((zerow (make-vector n 0)))
	      (let loop2 ()
		(if (< j n)
		    (let ((mrow (make-shared-array
				 m (lambda (a) (list j a)) n)))
		      (if (array-equal? mrow zerow)
			  (begin (set! j (+ j 1)) (loop2))))))
	      (if (< j n)
		  (let ((v (make-vector n 0))
			(mrowj (make-shared-array
				m (lambda (a) (list j a)) n)))
;		     (set! i (+ i 1))
;;;			 (ff:print " mrowj = " mrowj)
		    (array-copy! mrowj v)
;;;			 (ff:print " v = " v)
		    (set! ret (cons v ret))
		    (loop1 (+ j 1)))))))
;;;	  (ff:print " ret = " ret)
      (map (lambda (x) (ff:unorm (cons var (vector->list x)) p))
	   (reverse ret)))))

;;; return an ordered list of elements of Z mod n (symmetric)
(define (ff:generate-field n)
  (let ((v (make-vector n 1))
	(b (sym:bound n)))
    (array-index-map! v (lambda (i) (- i b)))
    (vector->list v)))

;;; return a sorted list of factors of a mod p, where a is square free
;;; and p is a prime number.
(define (ff:berlekamp a p)
  (let* ((q	(ff:q-matrix a p))
	 (n	(car (array-dimensions q)))
	 (var	(u:var a))
	 (qdiag (make-shared-array q (lambda (a) (list a a)) n)))
    (array-map! qdiag (lambda (x) (sym:sym (- x 1) p)) qdiag)
    (let* ((vs	(list->vector (ff:null-space-basis q (u:var a) p)))
	   (factors (list a))
	   (ffp (ff:generate-field p))
	   (k (vector-length vs)))
;;;	  (ff:print "Q = " q)
;;;	  (ff:print "vs are " vs)
      (let loop1 ((r 1))
	(if (< (length factors) k)
	    (begin
;;;		   (ff:print "loop1, r = " r "	vs = " vs)
	      (let loop2 ((us factors))
		(if (and (< (length factors) k) (not (null? us)))
		    (let ((u (car us))
			  (start #f))
;;;			 (ff:print " u = " u)
		      (let loop3 ((ss ffp))
			(if (not (null? ss))
			    (let ((s (car ss)))
;;;			       (ff:print " s = " s " " " r = " r " ")
			      (if (< (length factors) k)
				  (let ((g (ff:euclid-gcd
					    (ff:p-p (vector-ref vs r)
						    (list var s) p)
					    u p)))
;;; (ff:print (ff:p-p (vector-ref vs r) (list var s) p) (list var s) " g = " g)
				    (if (and
					 (not (equal? g u))
					 (not (u:one? g)))
					(begin
					  (set! factors (delete u factors))
					  (set! u (ff:p/p u g p))
					  (set! factors (delete
							 '()
							 (append factors
								 (list u)
								 (list g))))
					  (set! start #t)
;;; (ff:print " new factors = " factors)
					  )
					(loop3 (cdr ss))))))))
		      (loop2 (if start factors (cdr us))))))
	      (loop1 (+ r 1)))
	    (poly:sort-factors factors))))))

;;; Choose a prime number greater than and relatively prime to n1
(define (choose-prime n1)
  (choose-n (car *primes*) n1 *primes*))

(define (choose-n n n1 l)
  (if (and (> n n1) (one? (gcd n n1)))
      n
      (if (null? l)
	  (error "choose-n: Can't find a number larger than " n1)
	  (choose-n (car l) n1 (cdr l)))))

;;; hen:eea-lift computes s,t such that s a + t b = 1 (mod p^k),
;;; deg(s) < deg(b), deg(t) < deg(a), assuming GCD(a mod p, b mod p) =
;;; 1 in Z mod p [x], and that a and b have the same variable and are
;;; univariate.

;;; GCL Algorithm 6.3
;;; This is the extended euclidean gcd algorithm (mod p), but altered
;;; to lift its results to (mod p^k). Arguments a and b are assumed to
;;; be relatively prime.

(define hen:eea-lift
  (memon
   (lambda (a b p k)
     (if (not (equal? (u:var a) (u:var b)))
	 (error "In hen:eea-lift: polys a and b are not in the same variable."))
     (let* ((x (u:var a))
	    (amodp (ff:unorm a p))
	    (bmodp (ff:unorm b p))
	    (g (ff:eea-gcd amodp bmodp p))
	    (s (cadr g))
	    (smodp s)
	    (t (caddr g))
	    (tmodp t)
	    (modulus p))
;;;	  (ff:print "  a = " a "  b = " b)
       (do ((i 1 (+ i 1)))
	   ((>= i k) (let ((res (list s t)))
;;;			   (check-el a b p k res)
		       res))
;;; (ff:print "	 a = " a "  b = " b i s t modulus)
	 (let* ((e  (number->poly
		     (poly:- (list x 1) (poly:+ (poly:* s a) (poly:* t b)))
		     x))
		(c  (ff:unorm (u:p/c e modulus) p))
		(sb (number->poly (poly:* smodp c) x))
		(tb (number->poly (poly:* tmodp c) x))
		(qr (ff:p/p->qr sb bmodp p))
		(q  (car qr))
		(si (cadr qr))
		(ta (ff:p+p tb (ff:p*p q amodp p) p)))
;;;		(ff:print e c si ta)
	   (set! s (number->poly (poly:+ s (poly:* si modulus)) x))
	   (set! t (number->poly (poly:+ t (poly:* ta modulus)) x))
	   (set! modulus (* modulus p))))))))

(define (check-el a1 a2 p k result)
  (let* ((x (u:var a1))
	 (p^k (expt p k)))
    (let ((inner-prod (ff:unorm (number->poly
				(poly:+ (ff:p*p a1 (list-ref result 0) p^k)
					(ff:p*p a2 (list-ref result 1) p^k)) x)
			       p^k))
	  (cn (list x 1)))
;;;	  (ff:print " inner-prod = " inner-prod)
      (if (not (equal? cn inner-prod))
	  (begin
	    (ff:print "In hen:eea-lift: the result is not correct.")
	    (ff:print " a1 = " a1 " a2 = " a2 " result = " result)
	    (ff:print " inner-prod = " inner-prod
		      " supposed result = " cn " p^k = " p^k))))))


;;; Compute s1..sr such that s(1)*b(1)+...+s(r)*b(r) = 1 (mod p^k)
;;; with deg(s(j)) < deg(a(j))
;;; where b(i) = a(1)*...a(i-1)*a(i+1)*...*a(r),  i = 1..r
;;; Assumes p doesn't divide leading coefft of a(i), i = 1..r
;;;  and the a(i) must each be pairwise relatively prime in Z mod p [x].
;;; GCL alg 6.3
(define hen:multiterm-eea-lift
  (memon (lambda (a p k)
	   (let* ((r (length a))
		  (var (u:var (car a)))
		  (r-1 (- r 1))
		  (r-2 (- r 2))
		  (s	(make-vector r (list var 0)))
		  (q	(make-vector r-1 (list var 0)))
		  (beta (make-vector r (list var 1))))
	     (vector-set! q r-2 (list-ref a r-1))
	     (let loop ((j (- r-2 1)))
	       (if (>= j 0)
		   (begin
		     (vector-set! q j (poly:* (list-ref a (+ j 1))
					      (vector-ref q (+ j 1))))
		     (loop (- j 1)))))
;;;	  (ff:print " a = " a " q = " q)
	     (let loop ((j 0))
	       (if (< j r-1)
		   (let ((sigma (hen:diophantine
				 (list (vector-ref q j) (list-ref a j))
				 (vector-ref beta j) p k)))
		     (vector-set! beta (+ j 1) (car sigma))
		     (vector-set! s j	 (cadr sigma))
;;;		   (ff:print " beta = " beta " s = " s " sigma = " sigma)
		     (loop (+ j 1)))))
	     (vector-set! s r-1 (vector-ref beta r-1))
;;;	  (check-mel a p k (vector->list s))
	     (vector->list s)))))

(define (check-mel as p k result)
  (let* ((la (length as))
	 (x (u:var (car as)))
	 (p^k (expt p k))
	 (bs (make-vector la 0)))
;;;    (ff:print bs la)
    (let loop ((i 0))
      (if (< i la)
	  (begin
;;;		(ff:print "  i = " i)
	    (let loop1 ((j 0)
			(acc (list x 1)))
	      (if (< j la)
		  (begin
;;;			 (ff:print j bs)
		    (if (not (= i j))
			(loop1 (+ j 1) (ff:p*p acc (list-ref as j) p^k))
			(loop1 (+ j 1) acc)))
		  (vector-set! bs i acc)))
	    (loop (+ i 1)))))
    (set! bs (vector->list bs))
;;;    (ff:print "  bs = " bs)
    (let ((inner-prod (ff:unorm
		       (number->poly
			(reduce poly:+ (map (lambda (x y) (ff:p*p x y p^k))
					    bs result)) x) p^k))
	  (cn (list x 1)))
;;;	  (ff:print " inner-prod = " inner-prod " x^m = " x^m)
      (if (not (equal? cn inner-prod))
	  (begin
	    (ff:print "In hen:multiterm-eea-lift: the result is not correct.")
	    (ff:print " as = " as " bs = " bs " result = " result)
	    (ff:print " inner-prod = " inner-prod " supposed result = " cn
		   " p^k = " p^k))))))


;;; GCL alg 6.3
(define hen:diophantine1
  (memon (lambda (a x m p k)
	   (let ((r (length a))
		 (pk (expt p k))
		 (xm (poly:^ (list x 0 1) m)))
;;;    (ff:print "a = " a " x = " x " m = " m " pk = " pk " xm = " xm)
	     (if (> r 2)
		 (let ((s (hen:multiterm-eea-lift a p k))
		       (result '()))
		   (let loop ((j 0))
		     (if (< j r)
			 (begin
;;; (ff:print "s" j " = " (list-ref s j) " prod = " (poly:* xm (list-ref s j))
;;;	   " div = " (ff:p/p->qr (poly:* xm (list-ref s j))
;;;	    (list-ref a j) pk) " res = " result)
			   (set! result
				 (append
				  result
				  (cdr (ff:p/p->qr
					(poly:* xm (list-ref s j))
					(list-ref a j)
					pk))))
			   (loop (+ j 1)))
			 (begin
			   (check-hd1 a x m p k result)
			   result))))
		 (let* ((s (hen:eea-lift (list-ref a 1) (list-ref a 0) p k))
			(q1 (ff:p/p->qr (poly:* xm (list-ref s 0))
					(list-ref a 0)
					pk))
			(q (car q1))
			(result (list (cadr q1)
				      (ff:p+p (poly:* xm (list-ref s 1))
					      (poly:* q (list-ref a 1))
					      pk))))
		   (begin
;;;			(check-hd1 a x m p k result)
		     result)))))))

(define (check-hd1 as x m p k result)
  (let* ((la (length as))
	 (p^k (expt p k))
	 (bs (make-vector la 0)))
;;;    (ff:print bs la)
    (let loop ((i 0))
      (if (< i la)
	  (begin
;;;		(ff:print "  i = " i)
	    (let loop1 ((j 0)
			(acc (list x 1)))
	      (if (< j la)
		  (begin
;;;			 (ff:print j bs)
		    (if (not (= i j))
			(loop1 (+ j 1) (ff:p*p acc (list-ref as j) p^k))
			(loop1 (+ j 1) acc)))
		  (vector-set! bs i acc)))
	    (loop (+ i 1)))))
    (set! bs (vector->list bs))
;;;    (ff:print "  bs = " bs)
    (let ((inner-prod
	   (ff:unorm
	    (number->poly
	     (reduce poly:+ (map (lambda (x y) (ff:p*p x y p^k)) bs result))
	     x)
	    p^k))
	  (x^m (number->poly (poly:^ (list x 0 1) m) x)))
;;;	  (ff:print " inner-prod = " inner-prod " x^m = " x^m)
      (if (not (equal? x^m inner-prod))
	  (begin
	    (ff:print "In hen:diophantine1: the result is not correct.")
	    (ff:print " as = " as " bs = " bs " result = " result)
	    (ff:print " inner-prod = " inner-prod " x^m = " x^m " x = " x
		      " m = " m " p^k = " p^k))))))

;;; GCL alg 6.4 untested for most cases, works for two so far.
;;; Assumes that the leading coefficient of a given multivariate
;;; polynomial is the last one
;;;(load "ff")(trace hen:multivariate-diophantine hen:diophantine hen:mhenseln poly:evaln)

;;;(hen:mhenseln (list (sexp->var 'y) (list (sexp->var 'x) 0 0 1) (list (sexp->var 'x) 0 1))
;;;(list (list (sexp->var 'y) 1)) 3 1 (list (list (sexp->var 'x) 1 1) (list (sexp->var 'x) 0 1)) '(1 1))

;;;(hen:mhenseln (list (sexp->var 'y) 0 (list (sexp->var 'x) 0 1 1))
;;;(list (list (sexp->var 'y) 1)) 3 1 (list (list (sexp->var 'x) 1 1) (list (sexp->var 'x) 0 1)) '(1 1))

(define (hen:mhenseln a i p l u lcu)
  (let* ((nu-1 (length i))
	 (nu   (+ nu-1 1))
	 (pl   (expt p l))
	 (av   (make-vector nu a)))
(ff:print "a = " a " i = " i " u = " u)
    (let loop ((j (- nu 2)))
      (if (>= j 0)
	(let ((xj     (car (list-ref i j)))
	      (alphaj (cadr (list-ref i j))))
	  (vector-set! av j
	      (poly:evaln (vector-ref av (+ j 1)) xj alphaj))
;;;(ff:print "subst j = " j " xj = " xj " alphaj = " alphaj " av = " av)
	  (loop (- j 1)))))
    (let* ((maxdeg (apply max (map (lambda (x) (poly:degree a (car x))) i)))
	   (bu u)
	   (n  (length u)))
      (let loop ((j 0) (is i))
;;;(ff:print "j = " j " is = " is " nu-1 = " nu-1 " maxdeg = " maxdeg " bu = " bu " av = " av)
	(if (< j nu-1)
	  (let* ((bu1 bu)
		 (monomial 1)
		 (e 1)
		 (xj (car (list-ref i j)))
		 (alphaj (cadr (list-ref i j)))
		 (ij (list xj (- alphaj) 1)))
	    (set! bu
	      (map (lambda (x y)
		     (if (equal? y 1)
		       x
		       (append
			 (butlast x 1)
			 (list
			   (ff:mnorm
			     (reduce (lambda (z)
;;;(ff:print "y = " y " z = " z )
				       (poly:evaln y (car z) (cadr z))) is) pl)))))
		   bu lcu))
;;;(ff:print "bu = " bu " av(j+1) = " (vector-ref av (+ j 1)))
	    (set! e (poly:- (vector-ref av (+ j 1)) (reduce poly:* bu)))
(ff:print "e = " e)
	    (let loop1 ((k 1) (monomial ij))
(ff:print "k = " k " av(j+1) = " (poly:degree (vector-ref av (+ j 1)) xj))
	      (if (<= k (poly:degree (vector-ref av (+ j 1)) xj))
		(let ((c (taylor-coefficient e xj alphaj k)))
(ff:print "monomial = " monomial " c = " c)
(ff:print "bu1 = " bu1)
		  (if (not (equal? c 0))
		    (let* ((du1 (hen:multivariate-diophantine
				  bu1 c (butlast i (- nu j)) maxdeg p l))
			   (du (map (lambda (x) (poly:* (u:convert x) monomial)) du1)))
(ff:print "du1 = " du1)
(ff:print "du = " du)
;;;(ff:print "bu = " bu)
		      (set! bu (map (lambda (x y) (ff:mnorm (poly:+ x y) pl)) bu du))
(ff:print "bu = " bu)
		      (set! e (ff:mnorm (poly:- (vector-ref av (+ j 1))
					(reduce poly:* bu)) pl))))
(ff:print "e = " e)
		(loop1 (+ k 1) (poly:* monomial ij))))))))
;;;(ff:print "e = " e)
      (if (equal? a (reduce poly:* bu)) bu #f))))

;;; GCL alg 6.2 thoroughly untested except for two cases
(define (hen:multivariate-diophantine a c i d p k)
  (let* ((r	  (length a))
	 (nu-1	  (length i))
	 (nu	  (+ nu-1 1))
	 (pk	  (expt p k)))
(ff:print "r = " r " nu = " nu " pk = " pk " d = " d
	  " factors = " a " target = " c " i = " i)
    (if (> nu 1)
      (let* ((ap (reduce-init poly:* 1 a))
	     (b	 (map (lambda (x) (poly:/ ap x)) a))
	     (xnu     (car (list-ref i (- nu-1 1))))
	     (alphanu (cadr (list-ref i (- nu-1 1))))
	     (anew (poly:evaln a xnu alphanu))
	     (cnew (poly:evaln c xnu alphanu))
	     (inew (delete (list xnu alphanu) i))
;;;		(dummy (ff:print "b = " b " xnu = " xnu " alphanu = " alphanu
;;;				 " anew = " anew " cnew = " cnew " inew = " inew))
	     (sigma (hen:multivariate-diophantine anew cnew inew d p k))
	     (e (ff:mnorm
		  (poly:- c (reduce-init poly:+ 0 (map poly:* sigma b))) pk))
	     (monomial 1))
;;;(ff:print "b = " b " anew = " anew " cnew = " cnew " inew = " inew " sigma = " sigma
;;;	     " e = " e)
	 (let loop ((m 1))
	   (if (and (<= m d) (not (zero? e)))
	     (begin
	       (set! monomial (poly:* monomial (poly:- xnu alphanu)))
	       (let ((cm (taylor-coefficient e xnu m)))
;;;(ff:print "mononial = "monomial " cm = " cm)
		 (if (not (equal? cm 0))
		   (let ((ds hen:multivariate-diophantine anew cm inew d p k))
		     (set! ds (poly:* ds monomial))
		     (set! sigma (poly:* sigma ds))
		     (set! e (poly:-
			       (ff:p-p e
				       (reduce-init poly:+ 0
					  (map poly:* ds b)) pk)))
;;;(ff:print "ds = " ds " sigma = " sigma "e = " e)
		     ))))
	     (map (lambda (x) (ff:unorm x pk)) sigma ))))
      (let* ((x1      (u:var (car a)))
	     (sigma   (make-list r 0))
	     (dc      (poly:degree c x1)))
;;;(ff:print " a = " a " c = " c " sigma = " sigma)
	(let loop ((m 1))
;;;(ff:print " m = " m)
	  (if (<= (- m 1) dc)
	    (let* ((cm (leading-coeff (list-ref c m) x1))
		   (ds (hen:diophantine1 a x1 (- m 1) p k)))
	      (ff:print " cm = " cm " ds = " ds)
	      (set! ds (map (lambda (x) (u:p*c x cm)) ds))
	      (set! sigma (map (lambda (x y)
				 (number->poly (poly:+ x y) x1))
				 sigma ds))
	      (ff:print " sigma =" sigma " ds =" ds)
	      (loop (+ m 1)))
	    (map (lambda (x) (ff:unorm x pk)) sigma)))))))

;;	      (let ((result (map (lambda (x) (ff:unorm x p)) sigma)))
;;		(check-hd a c p k result)
;;		result)))))))

;;; Don't know if this one is OK, taken from poly.scm, commented out there.
(define (leading-coeff p var) (poly:coeff p var (poly:degree p var)))

;(define (poly:degree-me p var)
;   (cond ((number? p) 0)
;	  ((equal? var (car p)) (length (cddr p)))
;	  (else (reduce-init (lambda (m c) (max m (poly:degree-me c var)))
;			     0
;			     (cdr p)))))


;;; mth Taylor series coefficient in expansion about var = n
(define (taylor-coefficient poly var n m)
;;;  (ff:print "poly " poly "  var " var "  n " n "  m " m)
  (poly:/
    (poly:evaln (let loop ((i 1) (fn poly))
;;;		     (ff:print " i " i "  fn " fn)
		  (if (> i m) fn (loop (+ 1 i) (poly:diff fn var))))
		var n)
    (factorial m)))

;;;(define (factorial n)
;;;  (if (< n 2) 1 (* n (factorial (+ -1 n)))))

(define (factorial n)
  (letrec ((factorial1
	     (lambda (i acc)
	       (if (< i 2) acc (factorial1 (- i 1) (* acc i))))))
    (factorial1 n 1)))

;;; evaluate polynomials the naive way
(define (u:evaln p n)
  (let loop ((i 2) (i-1 1) (acc (list-ref p 1)))
    (if (> i (+ 1(u:degree p)))
      acc
      (loop (+ i 1) i (+ acc (* (list-ref p i) (expt  n i-1)))))))

(define (u:evalp p n)
  (let loop ((i 2) (i-1 1) (acc (list-ref p 1)))
    (if (> i (+ 1 (u:degree p)))
      acc
	(loop (+ i 1) i (poly:+ acc (poly:* (list-ref p i) (poly:^ n i-1)))))))

;;; checks if a polynomial is univariate
(define (poly:univariate? p)
  (every number? (cdr p)))

;;; evaluate a polynomial, p, substituting a number n for the variable var
(define (poly:evaln p var n)
;;;  (ff:print p var n)
  (cond
    ((number? p) p)
    ((equal? (car p) var)
      (if (poly:univariate? p)
	 (u:evaln p n)
	 (reduce poly:+
		(map (lambda (x y) (if (number? x)
				     (* x (expt n y))
				     (poly:* (poly:evaln x var n) (expt n y))))
		     (cdr p)
		     (power-list p var)))))
    (else
      (cons (car p)
	    (map (lambda (x) (if (number? x)
			       x (poly:evaln x var n)))
		 (cdr p))))))

;;; evaluate a polynomial p mod the ideal i
(define (poly:eval-ideal p i)
  (if (null? i) p (poly:eval-ideal (poly:evaln p (caar i) (cadar i)) (cdr i))))

;;; generate the powers of x for a poly
(define (power-list p v)
  (let ((vec (make-vector (+ 1 (poly:degree p v)) 1)))
    (array-index-map! vec (lambda (i) i))
    (vector->list vec)))

(define (index-list n)
  (let ((vec (make-vector n 1)))
    (array-index-map! vec (lambda (i) i))
    (vector->list vec)))

(define (index-vector n)
  (let ((vec (make-vector n 1)))
    (array-index-map! vec (lambda (i) i))
    vec))

;;; memoised diophantine equation solver is much faster
(define hen:diophantine
  (memon (lambda (a c p k)
	   (let* ((r	   (length a))
		  (p^k	   (expt p k))
		  (x1	   (u:var c))
		  (dc	   (u:degree c))
		  (sigma   (make-list r 0)))
;;;	  (ff:print " a = " a " c = " c " sigma = " sigma)
	     (let loop ((m 1))
;;;	     (ff:print " m = " m)
	       (if (<= (- m 1) dc)
		   (let* ((cm (list-ref c m))
			  (ds (hen:diophantine1 a x1 (- m 1) p k)))
;;;		   (ff:print " cm =" cm " ds =" ds)
		     (set! ds (map (lambda (x) (u:p*c x cm)) ds))
		     (set! sigma (map (lambda (x y)
					(number->poly (poly:+ x y) x1))
				      sigma ds))
;;;		   (ff:print " sigma =" sigma " ds =" ds)
		     (loop (+ m 1)))
		   (let ((result (map (lambda (x) (ff:unorm x p^k)) sigma)))
		     (check-hd a c p k result)
		     result)))))))

(define (check-hd as c p k result)
  (let* ((la (length as))
	 (x (u:var c))
	 (p^k (expt p k))
	 (bs (make-vector la 0)))
;;;    (ff:print bs la)
    (let loop ((i 0))
      (if (< i la)
	  (begin
;;;		(ff:print "  i = " i)
	    (let loop1 ((j 0)
			(acc (list x 1)))
	      (if (< j la)
		  (begin
;;;			 (ff:print j bs)
		    (if (not (= i j))
			(loop1 (+ j 1) (ff:p*p acc (list-ref as j) p^k))
			(loop1 (+ j 1) acc)))
		  (vector-set! bs i acc)))
	    (loop (+ i 1)))))
    (set! bs (vector->list bs))
;;;    (ff:print "  bs = " bs)
    (let ((inner-prod (ff:unorm
		       (number->poly
			(reduce poly:+ (map (lambda (x y) (ff:p*p x y p^k))
					    bs result)) x) p^k))
	  (cn (ff:unorm c p^k)))
;;;	  (ff:print " inner-prod = " inner-prod " x^m = " x^m)
      (if (not (equal? cn inner-prod))
	  (begin
	    (ff:print "In hen:diophantine: the result is not correct.")
	    (ff:print " as = " as " bs = " bs " result = " result)
	    (ff:print " inner-prod = " inner-prod " cn = " cn " p^k = " p^k)
	    )))))

;;; substitute new for old in x, preserving structure
(define (substitute new old x)
  (cond ((equal? new old) x)
	((equal? x old) new)
	((not (list? x)) x)
	((null? x) x)
	((null? (cdr x)) (list (substitute new old (car x))))
	(else (cons (substitute new old (car x))
		    (substitute new old (cdr x))))))

;;; 2 factor univariate linear Hensel lifting algorithm
(define (hen:ulhensel2 f g1 h1 p k)
  (let* ((g g1)
	 (h h1)
	 (lf (u:leading f))
	 (ghrecip (ff:eea-gcd g1 h1 p))
	 (grecip (cadr ghrecip))
	 (hrecip (caddr ghrecip))
	 (x (u:var f)))
;;;    (ff:print ghrecip)
    (let loop ((i 2)
	       (modulus p)
	       (modulus*p (* p p)))
;;;   (ff:print modulus modulus*p g h)
      (if (<= i k)
	  (let* ((discrepancy
		  (u:p/c (ff:p-p (ff:p/c f lf modulus*p)
				 (number->poly (poly:* g h) x) modulus*p)
			 modulus))
		 (gc (cadr (ff:p/p->qr
			    (number->poly (poly:* hrecip discrepancy) x) g1 p)))
		 (hc (cadr (ff:p/p->qr
			    (number->poly (poly:* grecip discrepancy) x)
			    h1
			    p))))
;;;		 (ff:print discrepancy gc hc)
	    (set! g (poly:+ g (poly:* gc modulus)))
	    (set! h (poly:+ h (poly:* hc modulus)))
;;;		 (ff:print i)
	    (loop (+ i 1) modulus*p (* modulus*p p)))
	  (list g h)))))

;;; Experimental (slow) 3 factor univariate linear Hensel lifting algorithm
(define (hen:ulhensel3 f g1 h1 i1 p k)
  (let* ((g g1)
	 (h h1)
	 (lf (u:leading f))
	 (i i1)
	 (x (u:var f)))
;;;    (ff:print g h i)
    (let loop ((l 2)
	       (modulus p)
	       (modulus*p (* p p)))
      (if (<= l k)
	  (let* ((diff
		  (u:p/c (ff:p-p (ff:p/c f lf modulus*p)
				 (number->poly (poly:* (poly:* g h) i) x)
				 modulus*p)
			 modulus))
		 (dghi (hen:diophantine (list g1 h1 i1) diff p 1))
;;;		     (dghi (hen:diophantine (list g1 h1 i1) diff p (- l 1)))
		 (dg (car dghi))
		 (dh (cadr dghi))
		 (di (caddr dghi)))
	    (set! g (poly:+ g (poly:* dg modulus)))
	    (set! h (poly:+ h (poly:* dh modulus)))
	    (set! i (poly:+ i (poly:* di modulus)))
;;;		(ff:print diff " dg = " dg " dh = " dh " di = " di g h i)
	    (loop (+ l 1) modulus*p (* modulus*p p)))
	  (list g h i)))))

;;; Experimental (slow) n factor univariate linear Hensel lifting algorithm
(define (hen:ulhenseln f as p k)
  (let ((x (u:var f))
	(lf (u:leading f)))
;;;    (ff:print f as)
    (let loop ((l 2)
	       (modulus p)
	       (modulus*p (* p p)))
      (if (<= l k)
	  (let* ((diff
		  (u:p/c (ff:p-p (ff:p/c f lf modulus*p) (number->poly (reduce poly:* as) x)
				 modulus*p) modulus))
		 (ds (hen:diophantine as diff p 1)))
;;;		 (ds (hen:diophantine as diff p (- l 1))))
	    (set! as (map (lambda (x y) (poly:+ x (poly:* y modulus))) as ds))
;;;	    (ff:print diff " ds = " ds " as = " as l)
	    (loop (+ l 1) modulus*p (* modulus*p p)))
	  as))))

;;; floored integer logarithm of x base b
(define (filog b x)
  (letrec ((log1 (lambda (b x acc)
		   (if (>= x b) (log1 b (quotient x b) (+ acc 1)) acc))))
    (log1 b x 0)))

;;; Return the Landau-Mignotte Bound of the coefficients of f
(define (landau-mignotte-bound f)
  (abs (* (sqrt (apply + (map (lambda (x) (* x x)) (cdr f))))
	  (expt 2 (u:degree f))
	  (u:leading f))))

;;; Experimental multivariate polynomial factoring
;;;(load "ff")(trace poly:factorsz u:factorsz hen:multivariate-diophantine hen:mhenseln)
(define (poly:factorsz poly)
  (trace poly:factorsz poly:factorszpp u:factorsz u:factorz
	 hen:multivariate-diophantine hen:mhenseln)
  (let* ((vars (poly:vars poly))
	 (nvars (length vars)))
    (cond
      ((= nvars 0) (list poly))				   ; number
      ((= nvars 1)					   ; univariate
	(let* ((slpoly (sign (u:leading poly)))
	       (cpoly  (u:cont poly))
	       (factors (poly:sort-factors (u:factorsz poly))))
	  (cond
	   ((and (= cpoly 1) (= slpoly 1))  factors)
	   (else (cons (list (u:var poly) (* slpoly cpoly)) factors)))))
      (else
	(let* ((degs (map (lambda (x) (poly:degree poly x)) vars))
	       (maxdeg (apply max degs))
	       (spec-var (list-ref vars (- nvars (length (memv maxdeg degs)))))
	       (pcont (poly:cont poly spec-var))
	       (ppoly (poly:pp poly spec-var)))
	  (ff:print "spec-var = " spec-var " pcont = " pcont "ppoly = " ppoly)
	  (cond
	    ((equal? pcont 1) (poly:sort-factors (poly:factorszpp ppoly)))
	    ((number? pcont) (cons (list pcont)
				   (poly:sort-factors
				     (poly:factorszpp ppoly))))
	    (else (poly:sort-factors (append (poly:factorszpp pcont)
					     (poly:factorszpp ppoly))))))))))

;;; Factorise a primitive multivariate polynomial poly
(define (poly:factorszpp ppoly)
  (let* ((vars (poly:vars ppoly))
	 (nvars (length vars))
	 (degs (map (lambda (x) (poly:degree ppoly x)) vars))
	 (maxdeg (apply max degs))
	 (spec-var (list-ref vars (- nvars (length (memv maxdeg degs)))))
	 (ivars (delete spec-var vars))
	 (p (let loop ((ns *primes*)
		       (cp (car *primes*)))
	      (if (not (= maxdeg (poly:degree (ff:mnorm ppoly cp) spec-var)))
		(begin
		  (ff:print "npoly = " (ff:mnorm ppoly cp)
			    " maxdeg = " maxdeg
			    " deg = " (poly:degree (ff:mnorm ppoly cp) spec-var))
		  (loop (cdr ns) (cadr ns)))
		 cp)))
	 (polymodp (ff:mnorm ppoly p))
	 (image ppoly)
	 (ideal (map
		  (lambda (x)
		    (let loop ((i 1))
		      (let* ((poly1 (poly:evaln image x i))
			     (is-sf (every (lambda (x) (eqv? x 1))
				      (map cadr (poly:sff poly1)))))
			(if (or (not (= maxdeg (poly:degree poly1 spec-var)))
				(not is-sf))
			  (begin
			    (ff:print "i " i "	p " p " poly1 " poly1
				      "	 x " x "  maxdeg " maxdeg
				      "	 deg " (poly:degree poly1 spec-var))
			    (loop (+ i 1)))
			  (begin
			    (set! image poly1)
			    (list x i))))))
		  ivars))
	 (image (poly:eval-ideal polymodp ideal))
	 (image-factors
	   (begin
	     (ff:print image " " p " " ideal " polymodp = " polymodp)
	     (let* ((slpoly (sign (u:leading image)))
		    (cpoly  (u:cont image))
		    (factors (poly:sort-factors (u:factorsz image))))
		 (cond
		   ((and (= cpoly 1) (= slpoly 1))  factors)
		   (else (cons (list spec-var (* slpoly cpoly)) factors)))))))
    (ff:print "image-factors = " image-factors)
    (if (= (length image-factors) 1)
      (list image)

;;; DANGER!! 2 used below for quick & dirty testing, must be replaced
;;; with a realistic bounding power. Likewise '(1 1) needs recursive
;;; leading coefficient factoring
;;; DANGER!! unsupported step here
      (hen:mhenseln ppoly ideal p 2 image-factors '(1 1 )))))


;		(lcsv (leading-coeff ppoly spec-var))
;		(svpoly (list spec-var 0 1))
;	     (if (equal? (poly:* svpoly lcsv) poly)
;	       (append (list svpoly) (poly:factorsz lcsv))

;;; Factorise a square-free univariate polynomial over the integers
(define (u:factorsz f1)
  (let* ((df (u:degree f1))
	 (var (u:var f1))
	 (lf  (u:leading f1))
	 (pfsp (let loop ((ns *primes*)
			  (p (car *primes*)))
;;;		    (ff:print var f1)
		 (if (or  (not (= df (ff:degree f1 p)))
			  (not (u:one? (ff:euclid-gcd (ff:diff f1 var p)
						      (ff:unorm f1 p) p))))
		     (loop (cdr ns) (cadr ns))
		     (list (ff:berlekamp (ff:make-monic f1 p) p) p))))
	 (pfs (car pfsp))
	 (p (cadr pfsp))
	 (n (length pfs))
	 (f (ff:make-monic f1 p))
	 (answer '()))
;;;    (ff:print "pfs = " pfs "answer = " answer "p = " p "n = " n)
;;;    (let ((fs pfs)
;;;	     (m n))
;;;	  (for-each (lambda (x)
;;;		       (if (eqv? 0 (cadr (poly:pdiv f x var)))
;;;			  (begin
;;;			     (set! answer (cons x answer))
;;;			     (set! pfs (delete	x pfs))
;;;			     (set! f (poly:/ f x))
;;;			     (set! m (- m 1)))))
;;;	      fs))
;;;    (ff:print pfs answer p n)
    (cond
     ((= n 1) (list f1))
     ((= n 2)
      (let* ((k (filog p (inexact->exact
			  (ceiling (* 2 (landau-mignotte-bound f1))))))
	     (factors (hen:ulhensel2 f1 (car pfs) (cadr pfs) p k))
	     (pk (expt p k))
	     (fs factors) (g '()))
	(for-each (lambda (x)
		    (if (eqv? 0 (cadr (poly:pdiv (poly:* f1 lf)
						 (ff:p*c x lf pk) var)))
			(begin
			  (set! answer (cons (u:pp (ff:p*c x lf pk)) answer))
			  (set! factors (delete	 x factors))
			  (set! n (- n 1)))))
		  fs)
;;;	(ff:print "k = " k " f = " f " f1 = " f1 " factors = " factors
;;;		  "  answer = " answer)
	(if (not (null? factors))
	    (list f1)
	    answer)))
     ((= n 3)
      (let* ((k (filog p (inexact->exact
			  (ceiling (* 2 (landau-mignotte-bound f1))))))
	     (factors (hen:ulhensel3 f1 (car pfs) (cadr pfs) (caddr pfs) p k))
	     (fs factors)
	     (pk (expt p k))
	     (g '()))
	(for-each (lambda (x)
		    (if (eqv? 0 (cadr (poly:pdiv (poly:* f1 lf)
						 (ff:p*c x lf pk) var)))
			(begin
			  (set! answer (cons (u:pp (ff:p*c x lf pk)) answer))
;;;				  (set! answer (cons x answer))
			  (set! factors (delete	 x factors))
			  (set! n (- n 1)))))
		  fs)
;;;		(ff:print factors answer)
	(if (not (null? factors))
	    (cons (u:pp (ff:p*c (ff:unorm (reduce-init
					  poly:*
					  (list var 1) factors) pk) lf pk))
		  answer)
	    answer)))
     (else
      (let* ((k (filog p (inexact->exact
			  (ceiling (* 2 (landau-mignotte-bound f1))))))
	     (factors (hen:ulhenseln f1 pfs p k))
	     (fs factors)
	     (pk (expt p k))
	     (e '())
	     (g '()))
	(for-each (lambda (x)
		    (if (eqv? 0 (cadr (poly:pdiv
				       (poly:* f1 lf) (ff:p*c x lf pk) var)))
			(begin
			  (set! answer (cons (u:pp (ff:p*c x lf pk)) answer))
;;;				  (set! answer (cons x answer))
			  (set! factors (delete	 x factors))
			  (set! n (- n 1)))))
		  fs)
;;;	(ff:print "n = " n " k = " k " factors = " factors "  answer = " answer)
	(call-with-current-continuation
	 (lambda (exit)
;;;		      (ff:print "at loop")
	   (let loop ((combine 2))
;;;			 (ff:print combine)
	     (if (<= (* 2 combine) n)
		 (let ((fs factors))
		   (set! e (combinations factors combine))
;;;			       (ff:print " e " e)
		   (for-each
		    (lambda (u)
		      (set! g (ff:unorm (reduce poly:* u) (expt p k)))
;;;		      (ff:print " g " g " fs " fs)
		      (if (eqv? 0 (cadr (poly:pdiv f1 (ff:p*c g lf pk) var)))
			  (begin
			    (set! answer (cons (u:pp (ff:p*c g lf pk)) answer))
;;;			    (set! answer (append (u:factorsz g) answer))
			    (for-each
			     (lambda (x)
			       (set! factors (delete x factors)))
			     u)
;;;			    (ff:print "factors = " factors "  answer = " answer)
			    (set! n (- n combine))
			    (if (> (* 2 combine) n) (exit #t)))))
		    e)
		   (loop (+ combine 1)))))))
	(if (not (null? factors))
	    (cons (u:pp (ff:p*c (reduce-init
				 poly:* (list var 1) factors) lf pk)) answer)
	    answer))))))

;;; Return all the unique subsets of size n obtainable from the list l
(define (combinations l n)
  (combs1 '() l n '()))

(define (combs1 l1 l2 n acc)
  (let* ((l1l (length l1))
	 (l2l (length l2))
	 (sumls (+ l1l l2l)))
    (cond
     ((< sumls n) acc)
     ((= sumls n) (append acc (list (append l1 l2))))
     ((= l1l (- n 1)) (append acc (combine-last l1 l2)))
     (else (apply append
		  (map (lambda (x y)
			 (combs1 (append l1 (list x)) y n acc))
		       l2 (make-ends (cdr l2) '())))))))
(define (make-ends l acc)
  (if (null? l)
      acc
      (make-ends (cdr l) (append acc (list l)))))

(define (list-head l n) (butlast l (- (length l) n)))

(define (combine-last lhead ltail)
  (map (lambda (x) (append lhead (list x))) ltail))

;;; return a sorted list of factors of the univariate polynomial poly
;;; over the integers (Z)
(define (u:factorz poly)
  (if (number? poly)
    (list (list poly 1))
    (let* ((slpoly (sign(u:leading poly)))
	   (cpoly  (u:cont poly))
	   (factors (map (lambda (x) (list (poly:sort-factors
					     (u:factorsz (car x)))
				     (cadr x)))
			 (poly:sff (poly:* slpoly (u:pp poly))))))
      (cond
       ((and (= cpoly 1) (= slpoly 1))	factors)
       (else (cons (list (list (* slpoly cpoly)) 1) factors))))))

;;; Least common multiple
(define (poly:lcm x)
   (poly:/ (reduce poly:* x)
	   (reduce poly:gcd x)))

;;; multiply powers of a list of factors l, by n
(define (mult-power l n)
   (map (lambda (x) (list (car x) (* n (cadr x)))) l))

;;; Factorise over the Rationals (Q)
(define (u:factorq poly)
  (cond
    ((rat? poly)
;;;	 (ff:print "poly is rat" poly)
      (append (u:factorz (num poly)) (mult-power (u:factorz (denom poly)) -1)))
    (else
;;;	 (ff:print "poly is not rational" poly)
      (u:factorz poly))))

;;; Idea copied from SLIB on Aubrey Jaffer's suggestion
(define (poly:sort-factors fs)
  (sort! fs poly:factor<))

(define (poly:factor< x y)
  (cond ((and (number? x) (number? y)) (< x y))
	((number? x) #t)
	((number? y) #f)
	((null? x) #t)
	((null? y) #f)
	((and (symbol? x) (symbol? y)) (string<? (symbol->string x)
						 (symbol->string y)))
	((vector? (car x))
	  (if (string<=? (vector-ref (car x) 1) (vector-ref (car y) 1))
	      (poly:factor< (cdr x) (cdr y)) #f))
	((> (length x) (length y)) #f)
	((< (length x) (length y)) #t)
	((and (list? x)
	      (list? y))
	   (cond
	     ((poly:factor< (car (last-pair x)) (car (last-pair y))) #t)
	     ((poly:factor< (car (last-pair y)) (car (last-pair x))) #f)
	     (else (poly:factor< (butlast x 1) (butlast y 1)))))
	((list? x)
	     (poly:factor< (but-last x 1) y))
	((list? y)
	     (poly:factor< x (but-last y 1)))
	(else
	     (slib:error "poly:factor<: unknown type" x y))))

;;; Convert a list of exponented factors to SEXP
;;; ((((x 1 1) (x 1 -1 1) (x 1 0 1) (x 7 0 1)) 1)) ->
;;; ((^ ( * (x 1 1) (x 1 -1 1) (x 1 0 1) (x 7 0 1)) 1))

;;; (define (factors->sexp fs)
;;;  (map (lambda (x) (insert^ (list (insert* (car x)) (cadr x)))) fs))

;;; (define (insert^ l) (cons '^ l))
;;; (define (insert* l) (cons '* l))

;;; Partial fraction expansion of a rational univariate polynomial
;;; The denominator, dr, must be square free.
(define (u:partial-fraction-expand nr dr)
  (let* ((drfs1 (u:factorq dr))
	 (drfs (remove-exponents drfs1 ()))
	 (p 3)
	 (k 5)
	 (ss   (hen:diophantine drfs (list (u:var (car drfs)) 1) p k)))
    (let ((res (map (lambda (x y) (make-rat (poly:* nr x) y)) ss drfs)))
      (ff:print res)
      res)))

;;;((#(:@ :@ #f () #f ()) (#(x x #f () #f ()) -4) (#(x x #f () #f ()) 1 -1))
;;;    (#(:@ :@ #f () #f ()) (#(x x #f () #f ()) 4) (#(x x #f () #f ()) -1 -1)))
;;;(k (filog p (inexact->exact (ceiling (* 2 (landau-mignotte-bound dr))))))

(define (remove-exponents fs fs1)
  (cond
    ((null? fs) fs1)
    ((number? (caar fs))
      (ff:print "n " (caar fs))
      (remove-exponents (cdr fs) (fs1)))
    (else
      (ff:print "e " (caar fs))
      (remove-exponents (cdr fs) (append (caar fs) fs1)))))

(define (ff:check-arg e)
  (cond ((not (u:univariate? e))
	 (bltn:error 'Not-a-univariate-polynomial poly))))

(define (ff:check-prime n)
  (cond ((not (memv n *primes*))
	 (bltn:error 'Not-a-Prime-Number n))))

;;; JACAL's builtin functions.
(defbltn 'sff
  (lambda (poly)
    (let ((e (licit->polxpr poly)))
      (ff:check-arg e)
      (cond
       ((not (equal? e (u:pp e)))	;this test should be replaced
	(bltn:error 'Not-a-primitive-polynomial poly))
       (else (poly:sff e))))))

(defbltn 'ffsff
  (lambda (poly pn . k1)
    (let ((p (licit->polxpr poly))
	  (n (licit->polxpr pn))
	  (k (licit->polxpr (if (null? k1) 1 (car k1)))))
      (ff:check-arg p)
      (ff:check-prime n)
      (cond
       ((not (ff:monic? p n))
	(bltn:error 'Not-monic-mod-n p))
       ((not (> k 0))
	(bltn:error 'Not-greater-than-zero k))
       (else (ff:sff p (expt n k)))))))

(defbltn 'berl
  (lambda (poly pn)
    (let ((p (licit->polxpr poly))
	  (n (licit->polxpr pn)))
      (ff:check-arg p)
      (ff:check-prime n)
      (cond
       ((not (ff:monic? p n))
	(bltn:error 'Not-monic-mod-n p))
       ((not (= (u:degree p) (ff:degree p n)))
	(bltn:error 'Not-same-degree-when-reduced-mod-n (ff:norm p n)))
       (else (ff:berlekamp p n))))))

(defbltn 'factoruz
  (lambda (poly)
;    (ff:print "poly = " poly)
    (let ((e1 (expr:normalize poly)))
;      (ff:print "e1 = " e1)
      (cond
	((rat? e1)
	  (bltn:error 'Rational-in-factoruz poly))
	((number? e1) (require 'factor) ;autoload from SLIB
		      (sort! (factor e1) <))
	(else (ff:check-arg e1) (u:factorz  e1))))))

(defbltn 'factoruq
  (lambda (poly)
    (let ((e1 (expr:normalize poly)))
      (cond
	((number? e1) (require 'factor) ;autoload from SLIB
		      (sort! (factor e1) <))
	(else (ff:check-arg e1)
	      (u:factorq e1))))))

(defbltn 'factormsz
  (lambda (poly)
    (let ((e1 (expr:normalize poly)))
      (cond
	((number? e1) (require 'factor) ;autoload from SLIB
		      (sort! (factor e1) <))
	((not (list?  e1))
	  (bltn:error 'Not-a-polynomial poly))
	((not (poly:var? (car e1)))
	  (bltn:error 'Not-a-polynomial poly))
	(else
	  (poly:factorsz e1))))))

(defbltn 'parfrac
  (lambda (poly)
    (let ((e1 (expr:normalize poly)))
      (u:partial-fraction-expand (num e1) (denom e1)))))

;;; (x+1)/(x+2)
;;;(#(:@ :@ #f () #f ()) (#(x x #f () #f ()) -1 -1) (#(x x #f () #f ()) 2 1))

(define a '(x -5 2 8 -3 -3 0 1 0 1))
(define b '(x 21 -9 -4 0 5 0 3))
(define b1 '(x -2 -9 -4 0 5 0 3))
(define c '(x 1 1 2 2 1 1))
(define d '(x 1 0 2 0 1))
(define e '(x 2 1 4 2 2 1))
(define f '(x -36 42 -84 48))
(define g '(x -30 44 -10 -4))
(define h '(x -5 2 1 -3 -3 0 1 0 1))
(define i '(x 21 -9 -4 0 5 0 3))

(define j '(x -5 2 8 -3 -3 0 1 0 1))	; page 280
(define k '(x -2 -9 -4 0 5 0 3))	; page 280 gcd 21 mod 23

(define l '(x -360 -171 145 25 1))
(define m '(x -15 -14 -1 15 14 1))	; page 302 mod 5 = x^4 + 4x
					;	   mod 7 = x^2 + 1
					;	   mod 11 = x^2 + 3x + 4
(define o '(x 1 0 2 2 0 1 1 0 2 2 0 1))
(define p (ff:diff o 'x 3))

(define q '(x -1 0 2 0 0 0 -2 0 1))

(define m1 '#2(( 0   0	 0   0	 0   0)
	       ( 3   4	-3  -3	-5   5)
	       ( 3  -5	 5   1	-1   0)
	       (-2   4	-1   2	-4  -2)
	       (-4  -3	-1   0	-1  -3)
	       (-3  -1	-4  -3	-1  -4)))

(define m2 '#2((0  0  0)
	       (1  2  3)
	       (4  5  6)))

(define m3 '#2((0  0  0	 0  0  0  0  0)
	       (2  0  7 11 10 12  5 11)
	       (3  6  3	 3  0  4  7  2)
	       (4  3  6	 4  1  6  2  3)
	       (2 11  8	 8  2  1  3 11)
	       (6 11  8	 6  2  6 10  9)
	       (5 11  7 10  0 11  6 12)
	       (3  3 12	 5  0 11  9 11)))


(define p1 '(x 1 2 1))
(define p2 '(x 1 3 3 1))
(define p3 '(x 1 4 6 4 1))
(define p4 '(x 1 -2 1))
(define p5 '(x 1 0 1))
(define p6 '(x -1 0 1))
(define p7 '(x 1 -3 -1 -3 1 -3 1))
(define p8 '(x 8 2 8 10 10 0 1 0 1))
(define p9 '(x 4 12 13 6 1))
(define p10 '(x 1 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 1)) ; x^20+1
(define p11 '(x 5040 -432 10 1))	; GCL p234
(define p12 '(x 6 11 6 1))		; (x+1)(x+2)(x+3)
(define p13 '(x 7 0 8 7 1 8 0 1))	; (x+1)(x^2-x+1)(x^2+1)(x^2+7)
(define p14 '(x -1000 3650 10720 1415 -8198 -8235 -11288 -1894 9244
		4584 568 479 -46 1))
(define mv1 (list (sexp->var 'z) (list (sexp->var 'x) 0 2)  1 (list (sexp->var 'x) 0 0 1)))
(define mv2 (list (sexp->var 'z) (list (sexp->var 'x) 0 1)  1 1))
(define mv3 (list (sexp->var 'z) (list (sexp->var 'x) 10 7)  1 -8))
(define mv4 (list (sexp->var 'z) (list (sexp->var 'x) 0 2)  (list (sexp->var 'x) 0 0 1)))

; test data  (Amiga 1200 2Mb Chip RAM 4Mb Fast RAM 50MHz 68030 CPU No FPU WB 3.0)
; ff.scm 1  Jan 1995   28460 msec (8420 msec gc) 653689 cells work (62811 cells other)
; ff.scm 27 Jan 1995   27800 msec (7960 msec gc) 653681 cells work (62811 cells other)
;		       (using = instead of eqv?)
;		       25640	   7600		 573246		    62811
;		       (and speeding up ff:leading)
;		       24760	   7040		 572511		    62811
;		       (and using cadr in u:one? u:zero? etc)
;		       22260	   6380		 487313		    62811
;		       (and removing ff:unorm from ff:pp ff:cont)
;		       19340	   4960		 283401		   126751
;		       (and speeding up ff:unorm)
;		       15500	   3920		 241554		   100271
;		       (and using u:lose-lzeros)
;		       12160	   3020		 226016		    62811
;		       (and rewriting ff:unorm)
;		       11320	   2920		 222492		    45683
;		       (and rewriting ff:p*c)
;		       10000	   2740		 203486		    30979
;		       (and rewriting ff:p+p, ff:p-p)
;			9960	   2780		 197025		    30979
;		       (and alter ff:p+p, ff:p-p, ff:p/c)
;			9500	   2700		 191548		    19235
;			9080	   2280		 189661		    19235
;		       (and alter ff:p/p)
;   30 January 1995    17520	   4680		 377861		    19831
;		       (and went over to symmetric representation)
;   31 January 1995    17300	   4820		 367493		    19679
;		       (and rewrote ff:p/p, using new ff:p/p->qr)
;   1 February 1995    11220	   3160		 218731		    23131
;		       (and used ff:p/p->qr in ff:euclid-gcd)
;   7 February 1995    11560	   3340		 211306		    24035
;  17 February 1995    10940	   2900		 192896		    24035
;  23 February 1995	9880	   3320		 157728		    28139
;		       (using latest scm 4e2)
;   1 March    1995	9403	   2852		 152970		    28139
;		       (scm 4e2 with CCLO's, and added result sorting)
;

(define (ff:test)
  (test '(x 1)
	ff:euclid-gcd j k 23)
  (test '(x 0 -1 0 0 1)
	ff:euclid-gcd l m 5)
  (test '(x 1 0 1)
	ff:euclid-gcd l m 7)
  (test '(x 4 3 1)
	ff:euclid-gcd l m 11)
  (test '((x -4 1) (x 4 1) (x 5 0 1))
	ff:berlekamp '(x 8 0 0 0 1) 11)
  (test '((x 1 1) (x 3 5 1) (x 4 3 2 1) )
	ff:berlekamp p7 11)
  (test '((x 3 1) (x -1 4 -5 1) (x 6 4 3 2 1))
	ff:berlekamp p8 13)
  (test '(((x 1 1) 2))
	poly:sff p1)
  (test '(((x -1 1) 2))
	poly:sff p4)
  (test '(((x 2 3 1) 2))
	poly:sff p9)
  (test '(((x 8 2 8 10 10 0 1 0 1) 1))
	poly:sff p8)
  'done)

;;; ff:test2
;;;   4 February  1995	13260	    3360	  246229	       62499
;;;   1 March	  1995	13618	    4042	  208548	       63807
;;;			(scm 4e2 with CCLO's, and added result sorting)

(define (ff:test2)
  (test '((x -1 -1 1) (x -1 1 1) (x 1 0 1 -1 1) (x 1 -1 1 0 1) (x 1 1 1 0 1) (x 1 0 1 1 1))
	 ff:berlekamp p10 3)
  'done)

;;; ff:test3
;;;   4 February  1995	32100	    8020	  602668	       81739
;;;  23 February  1995	29840	    8440	  460900	       84943
;;;			(using latest scm 4e2)
;;;   1 March	  1995	29706	    8840	  454745	       84943
;;;			(scm 4e2 with CCLO's, and added result sorting)

(define (ff:test3)
  (test '((x -5 0 1) (x 5 0 1) (x -1 1 2 -5 1) (x -1 5 -2 -1 1) (x -1 -5 -2 1 1) (x -1 -1 2 5 1))
	 ff:berlekamp p10 13)
  'done)


;;; ff:test4
;;;  21 February  1995	74360 mSec (21780 in gc) 1391234 cells work, 59617 bytes other
;;;			29820	     8540	  539821	     24317
;;;			(after memoisation of hen:diophantine. 4 seconds on second run)
;;;			25980	     7560	  460937	     20637
;;;			(after memoization of hen:eea-lift etc)
;;;  22 February 1995	22040	     3660	  459685	     19653
;;;			(after removal of some un-needed error checking)
;;;  23 February 1995	26820	     8680	  466699	     21261
;;;   1 March	 1995	26462	     8463	  456924	     22793
;;;			(scm 4e2 with CCLO's, and added result sorting)

(define (ff:test4)
  (test '((((x 1 1) (x 1 -1 1) (x 1 0 1) (x 7 0 1)) 1))
	u:factorz p13)
  'done)
